/*
	Copyright 2009 Jerry Ylilammi
*/

#pragma once

#include "maxscrpt/MAXScrpt.h"

#include "field.h"

#include <algorithm>
#include <vector>

/*
	In static version there is no need to calculate point positions
	for each ray. So for container we'll just use the leaf.
*/
class BVHContainer {
public:
	virtual void getFieldValueNoTest(FieldValue* field) = 0;
};

/*
	When we trace the BVH each leaf thats along the ray is turned into these
	They contain the positions along the ray where it enters the bounding box
	of the leaf and when it exits it. This means the BVH has to be intersected
	for each ray only once.
*/
class BVHStep {
public:
	VR::Ireal min;
	VR::Ireal max;
	BVHContainer* cont;
	BVHStep* next;
	
	BVHStep(VR::Ireal min, VR::Ireal max, BVHContainer* cont, BVHStep* next) {
		this->min=min; this->max=max; this->cont=cont; this->next=next;
	}
	~BVHStep() {delete next;}
	void deleteThisOnly() {next = NULL; delete this;} // Used when optimizing, sometimes we want to remove individual steps
};
struct BVHStepSort {
     bool operator()(BVHStep*& a, BVHStep*& b) {
          return a->min < b->min;
     }
};

/*
	Ray structure that contains all the leafs along single ray
*/
class BVHRay {	
public:
	BVHStep* first;	
	std::vector<BVHStep*> steps;

	void add(VR::Ireal min, VR::Ireal max, BVHContainer* cont);	
	void sortSteps();
	void collapseSteps();

	BVHRay() {first = NULL; steps.reserve(25); }
	~BVHRay() {delete first;}
};

/*
	Parent class for BVH nodes
*/
class BVHNode {
public:
	VR::Box b;
		
	BVHNode() {}
	~BVHNode() {}

	int intersect(VR::IRay* ray);

	// Override
	virtual int type() {return 0;}
	virtual void traceRay(VR::IRay* ray, BVHRay* bRay) = 0;
	virtual void getFieldValue(FieldValue* field) = 0;
	virtual void getGradValue(FieldGrad* grad) = 0;

	// For branches
	virtual BVHNode* getLeft() {return NULL;}
	virtual BVHNode* getRight() {return NULL;}

	// For leaves
	virtual int* getPoints() {return NULL;}	
	virtual int getCount() {return 0;}
};

/*
	BVH branch class
*/
class Branch: public BVHNode {
public:
	BVHNode* left;
	BVHNode* right;

	int type() {return 1;}			
	void traceRay(VR::IRay* ray, BVHRay* bRay);
	void getFieldValue(FieldValue* field);
	void getGradValue(FieldGrad* grad);

	BVHNode* getLeft() {return left;}
	BVHNode* getRight() {return right;}

	Branch(VR::Vector* pointCloud, VR::Ireal* radii, int* points, int count, VR::real size, int maxDepth, int depth, int leafSize, VR::real maxLength);	
	~Branch() {
		delete left;
		delete right;
	}
};


/*
	BVH leaf class, leafs only store indexes of points, they don't
	store the actual points.
*/
class Leaf: public BVHNode, public BVHContainer {
public:
	int* points;
	int count;

	int type() {return 0;}	
	void traceRay(VR::IRay* ray, BVHRay* bRay);
	void getFieldValue(FieldValue* field);
	void getFieldValueNoTest(FieldValue* field);
	void getGradValue(FieldGrad* grad);

	int* getPoints() {return points;}
	int getCount() {return count;}	
	
	Leaf(VR::Vector* pointCloud, VR::Ireal* radii, int* points, int count, VR::real size);	
	~Leaf() {
		delete[] points;
	}
};

/*
	BVH base class
*/
class BoundingVolumeHierarchy {
private:
	BVHNode* root;
	
public:
	BoundingVolumeHierarchy() {
		root = NULL;
	}

	~BoundingVolumeHierarchy() {
		delete root;
	}

	void create(VR::Vector* pointCloud, VR::Ireal* radii, int count, VR::real size, int maxDepth, int leafSize, VR::real maxLength);
	int intersect(VR::IRay* ray) {	
		return root->intersect(ray);
	}
	void traceRay(VR::IRay* ray, BVHRay* bRay) {
		root->traceRay(ray, bRay);
		bRay->sortSteps();
		bRay->collapseSteps();			
	}
	void getFieldValue(FieldValue* field) {
		root->getFieldValue(field);
	}
	void getGradValue(FieldGrad* grad) {
		root->getGradValue(grad);
	}
};