#include "max.h"
#include "BoundingVolumeHierarchy.h"

using namespace VR;

static void bAddExpanded(Box &b, const Vector &p, real dist) {
	for (int i=0; i<3; i++) {
		if (p[i]-dist<b.pmin[i]) b.pmin[i]=p[i]-dist;
		if (p[i]+dist>b.pmax[i]) b.pmax[i]=p[i]+dist;
	}
}

/*
	############ BVHRay ############
*/

// TODO Optimize search, now brute force
void BVHRay::add(VR::Ireal min, VR::Ireal max, BVHContainer* cont) {
	//if (min < 0.) min = 0.;

	if (this->first == NULL) {
		this->first = new BVHStep(min, max, cont, NULL);
		return;
	}

	if (min < this->first->min) {
		this->first = new BVHStep(min, max, cont, this->first);
		return;
	}

	BVHStep* cur = this->first;
	while (cur->next != NULL && min > cur->next->min)
		cur = cur->next;
	cur->next = new BVHStep(min, max, cont, cur->next);
}

/*
	############ BVHNode ############
*/

int BVHNode::intersect(VR::IRay* ray) {
	if (ray->testIntersectBox(this->b)) {
		if (!this->type())
			return true;
		if (this->getLeft()->intersect(ray))
			return true;
		if (this->getRight()->intersect(ray))
			return true;
	}	
	return false;
}

/*
	############ Branch ############
*/

void Branch::traceRay(VR::IRay* ray, BVHRay* bRay) {
	if (ray->testIntersectBox(this->b)) {
		this->getLeft()->traceRay(ray, bRay);
		this->getRight()->traceRay(ray, bRay);
	}
}

void Branch::traceRayInside(VR::IRay* ray, BVHRay* bRay) {
	//IRay ray2 = IRay(*ray);
	if (ray->testIntersectBox(this->b)) {
		this->getLeft()->traceRayInside(ray, bRay);
		this->getRight()->traceRayInside(ray, bRay);
	}
}

void Branch::getFieldValue(VR::Vector* pointCloud, VR::Vector location, Ireal size, Ireal size2, Ireal (*function)(Ireal, Ireal), real* res) {
	if (b.isInside(location)) {
		this->getLeft()->getFieldValue(pointCloud, location, size, size2, function, res);
		this->getRight()->getFieldValue(pointCloud, location, size, size2, function, res);
	}
}

void Branch::getGradValue(VR::Vector* pointCloud, VR::Vector location[4], Ireal size, Ireal size2, Ireal (*function)(Ireal, Ireal), Ireal f[4]) {
	if (b.isInside(location[0])) {
		this->getLeft()->getGradValue(pointCloud, location, size, size2, function, f);
		this->getRight()->getGradValue(pointCloud, location, size, size2, function, f);
	}
}

Branch::Branch(VR::Vector* pointCloud, int* points, int count, real size, int maxDepth, int depth) {
	// Create bounding box
	b.init();
	for (int i=0;i<count;i++)
		bAddExpanded(b, pointCloud[points[i]], size);

	// Choose which axis to split and where
	int axis = b.maxDimension();
	real middle = (b.pmin[axis] + b.pmax[axis]) * 0.5f;
	
	// Count number of points on both sides
	int leftCount = 0;
	for (int i=0;i<count;i++)
		if (pointCloud[points[i]][axis] < middle)
			leftCount++;
	int rightCount = count - leftCount;

	// If the other side gets 0 points, all the points are at the same location
	// and there is no longer point dividing them any futher so we just make
	// leaves here
	bool forceLeaf = false;
	if (leftCount == 0 || rightCount == 0) {
		leftCount = count / 2;
		rightCount = count - leftCount;
		forceLeaf = true;
	}

	// Create fitting arrays for both
	int* leftPoints;
	int* rightPoints;	
	//if (leftCount > 0)
		leftPoints = new int[leftCount];
	//if (rightCount > 0)
		rightPoints = new int[rightCount];
	
	// Assign points to for both sides
	int j = 0, k = 0;
	for (int i=0;i<count;i++)
		if (pointCloud[points[i]][axis] < middle) {
			if (j < leftCount)
				leftPoints[j++] = points[i];
		} else {
			if (k < rightCount)
				rightPoints[k++] = points[i];
		}

	// Free previous point array, as its now split into left and right
	delete[] points;
	
	if (maxDepth <= depth || leftCount < 3 || forceLeaf) // Make leaf
		this->left = new Leaf(pointCloud, leftPoints, leftCount, size);
	else // Branch more		
		this->left = new Branch(pointCloud, leftPoints, leftCount, size, maxDepth, depth+1);

	if (maxDepth <= depth || rightCount < 3 || forceLeaf) // Make leaf
		this->right = new Leaf(pointCloud, rightPoints, rightCount, size);
	else // Branch more		
		this->right = new Branch(pointCloud, rightPoints, rightCount, size, maxDepth, depth+1);
}

/*
	############ Leaf ############
*/

void Leaf::traceRay(VR::IRay* ray, BVHRay* bRay) {
	IRay ray2 = IRay(*ray);
	if (ray2.intersectBox(this->b))
		bRay->add(ray2.cmint, ray2.cmaxt, this);			
}

void Leaf::traceRayInside(VR::IRay* ray, BVHRay* bRay) {
	IRay ray2 = IRay(*ray);
	if (ray2.intersectBox(this->b))
		bRay->add(ray2.cmint, ray2.cmaxt, this);			
}

void Leaf::getFieldValue(VR::Vector* pointCloud, VR::Vector location, Ireal size, Ireal size2, Ireal (*function)(Ireal, Ireal), real* res) {
	if (b.isInside(location)) {
		Ireal r;
		for (int i=0;i<this->count;i++) {
			r = (pointCloud[points[i]]-location).lengthSqr();
			if (r > size2) continue;
			*res += function(size2, r);
		}
	}
}

void Leaf::getGradValue(VR::Vector* pointCloud, VR::Vector location[4], Ireal size, Ireal size2, Ireal (*function)(Ireal, Ireal), Ireal f[4]) {
	if (b.isInside(location[0])) {
		Ireal r;
		for (int i=0;i<this->count;i++) {			
			for (int j=0;j<4;j++) {
				r = (pointCloud[points[i]]-location[j]).lengthSqr();
				if (r > size2) continue;
				f[j] += function(size2, r);
			}
		}
	}
}

Leaf::Leaf(VR::Vector* pointCloud, int* points, int count, real size) {
	b.init();
	for (int i=0;i<count;i++)
		bAddExpanded(b, pointCloud[points[i]], size);
	this->points = points;
	this->count = count;		
}

/*
	############ BoundingVolumeHierarchy ############
*/

void BoundingVolumeHierarchy::create(VR::Vector* pointCloud, int count, real size, int maxDepth) {
	delete this->root;
	int* points = new int[count];
	for (int i=0;i<count;i++)
		points[i] = i;
	this->root = new Branch(pointCloud, points, count, size, maxDepth, 1);
}

#define NORMALDELTA 0.001f
Vector BoundingVolumeHierarchy::getFieldGrad(VR::Vector* pointCloud, VR::Vector location, Ireal size, Ireal size2, Ireal (*function)(Ireal, Ireal)) {
	Ireal f[4];
	for (int i=0;i<4;i++)
		f[i] = 0.f;
	Vector loc[4];
	loc[0] = location;
	loc[1] = location - Vector(NORMALDELTA,0.f,0.f);
	loc[2] = location - Vector(0.f,NORMALDELTA,0.f);
	loc[3] = location - Vector(0.f,0.f,NORMALDELTA);
	this->root->getGradValue(pointCloud, loc, size, size2, function, f);
	f[1] = f[0] - f[1];
	f[2] = f[0] - f[2];
	f[3] = f[0] - f[3];
	Ireal length = sqrt(f[1]*f[1] + f[2]*f[2] + f[3]*f[3]);
	f[1] /= length;
	f[2] /= length;
	f[3] /= length;
	return -Vector(f[1], f[2], f[3]);	
}