#include "vraybase.h"
#include "vraygeom.h"
#include "rayserver.h"

#include "vrayrenderer.h"
#include "vrayplugins.h"
#include "vraytexutils.h"

#include "maxscrpt/MAXScrpt.h"

class BVHContainer {
public:
	virtual void getFieldValue(VR::Vector* pointCloud, VR::Vector location, VR::Ireal size, VR::Ireal size2, VR::Ireal (*function)(VR::Ireal, VR::Ireal), VR::real* res) = 0;
};

class BVHStep {
public:
	VR::Ireal min;
	VR::Ireal max;
	BVHContainer* cont;
	BVHStep* next;

	void deleteThisOnly() {next = NULL; delete this;}
	
	BVHStep(VR::Ireal min, VR::Ireal max, BVHContainer* cont, BVHStep* next) {
		this->min=min; this->max=max; this->cont=cont; this->next=next;
	}
	~BVHStep() {delete next;}
};

class BVHRay {
public:
	BVHStep* first;

	void add(VR::Ireal min, VR::Ireal max, BVHContainer* cont);
	
	BVHRay() {first = NULL;}
	~BVHRay() {delete first;}
};

class BVHNode: public BVHContainer {
public:
	VR::Box b;
		
	BVHNode() {}
	~BVHNode() {}

	int intersect(VR::IRay* ray);	

	void getBranchCount(int& count) {
		if (this->type() == 0) return;
		count++;
		this->getLeft()->getBranchCount(count);
		this->getRight()->getBranchCount(count);
	}
	void getLeafCount(int& count) {
		if (this->type() == 0) {			
			count++;
			return;
		}		
		this->getLeft()->getLeafCount(count);
		this->getRight()->getLeafCount(count);
	}

	// Override
	virtual int type() {return 0;}
	virtual void traceRay(VR::IRay* ray, BVHRay* bRay) = 0;
	virtual void traceRayInside(VR::IRay* ray, BVHRay* bRay) = 0;	
	virtual void getFieldValue(VR::Vector* pointCloud, VR::Vector location, VR::Ireal size, VR::Ireal size2, VR::Ireal (*function)(VR::Ireal, VR::Ireal), VR::real* res) = 0;
	virtual void getGradValue(VR::Vector* pointCloud, VR::Vector location[4], VR::Ireal size, VR::Ireal size2, VR::Ireal (*function)(VR::Ireal, VR::Ireal), VR::Ireal f[4]) = 0;

	// For branches
	virtual BVHNode* getLeft() {return NULL;}
	virtual BVHNode* getRight() {return NULL;}

	// For leaves
	virtual int* getPoints() {return NULL;}	
	virtual int getCount() {return 0;}
};

class Branch: public BVHNode {
public:
	BVHNode* left;
	BVHNode* right;

	int type() {return 1;}			
	void traceRay(VR::IRay* ray, BVHRay* bRay);
	void traceRayInside(VR::IRay* ray, BVHRay* bRay);
	void getFieldValue(VR::Vector* pointCloud, VR::Vector location, VR::Ireal size, VR::Ireal size2, VR::Ireal (*function)(VR::Ireal, VR::Ireal), VR::real* res);
	void getGradValue(VR::Vector* pointCloud, VR::Vector location[4], VR::Ireal size, VR::Ireal size2, VR::Ireal (*function)(VR::Ireal, VR::Ireal), VR::Ireal f[4]);
	BVHNode* getLeft() {return left;}
	BVHNode* getRight() {return right;}

	Branch(VR::Vector* pointCloud, int* points, int count, VR::real size, int maxDepth, int depth);	
	~Branch() {
		delete left;
		delete right;
	}
};

class Leaf: public BVHNode {
public:
	int* points;
	int count;

	int type() {return 0;}	
	void traceRay(VR::IRay* ray, BVHRay* bRay);
	void traceRayInside(VR::IRay* ray, BVHRay* bRay);
	void getFieldValue(VR::Vector* pointCloud, VR::Vector location, VR::Ireal size, VR::Ireal size2, VR::Ireal (*function)(VR::Ireal, VR::Ireal), VR::real* res);
	void getGradValue(VR::Vector* pointCloud, VR::Vector location[4], VR::Ireal size, VR::Ireal size2, VR::Ireal (*function)(VR::Ireal, VR::Ireal), VR::Ireal f[4]);
	int* getPoints() {return points;}
	int getCount() {return count;}	
	
	Leaf(VR::Vector* pointCloud, int* points, int count, VR::real size);	
	~Leaf() {
		delete[] points;
	}
};

class BoundingVolumeHierarchy {
private:
	BVHNode* root;
	
public:
	BoundingVolumeHierarchy() {
		root = NULL;
	}

	~BoundingVolumeHierarchy() {
		delete root;
	}

	void create(VR::Vector* pointCloud, int count, VR::real size, int maxDepth);
	int intersect(VR::IRay* ray) {	
		return root->intersect(ray);
	}
	void traceRay(VR::IRay* ray, BVHRay* bRay) {
		root->traceRay(ray, bRay);
	}
	void traceRayInside(VR::IRay* ray, BVHRay* bRay) {
		root->traceRayInside(ray, bRay);
	}

	int getBrachCount() { int count = 0; root->getBranchCount(count); return count;}
	int getLeafCount() { int count = 0; root->getLeafCount(count); return count;}
	void getFieldValue(VR::Vector* pointCloud, VR::Vector location, VR::Ireal size, VR::Ireal size2, VR::Ireal (*function)(VR::Ireal, VR::Ireal), VR::real* res) {
		root->getFieldValue(pointCloud, location, size, size2, function, res);
	}
	VR::Vector getFieldGrad(VR::Vector* pointCloud, VR::Vector location, VR::Ireal size, VR::Ireal size2, VR::Ireal (*function)(VR::Ireal, VR::Ireal));
};