/*
	Copyright 2009 Jerry Ylilammi
*/

#pragma once

#include "vraybase.h"

/*
	These classes do the actual field computation.

	Field* class is passed around and each BVH container adds its points to it with addPoints()

	This way we can implement any kind of technique to compute the field.

	Good example of this is the average field method where weighted average of all the points locations
	is used.
*/


/*
	Normal field value computations, field = sum of all metaballs
*/

class FieldValue {
public:	
	VR::Ireal res;
	VR::Vector location;
	VR::Ireal size2;	
	VR::Ireal (*function)(VR::Ireal, VR::Ireal);
	VR::Vector* pointCloud;	
	VR::Ireal* radii;
	
	virtual void addPoints(int* points, int count);
	FieldValue() {}
	FieldValue(VR::Vector* pointCloud, VR::Ireal* radii, VR::Vector location, VR::Ireal (*function)(VR::Ireal, VR::Ireal), VR::real size2) {
		this->pointCloud=pointCloud; this->radii=radii; this->location=location; this->function=function; this->size2=size2; res=0.;
	}
	~FieldValue() {}
};

class FieldGrad {
public:
	VR::Ireal res[4];
	VR::Vector location[4];
	VR::Ireal size2;	
	VR::Ireal (*function)(VR::Ireal, VR::Ireal);
	VR::Vector* pointCloud;
	VR::Ireal* radii;

	virtual void addPoints(int* points, int count);
	FieldGrad() {for (int i=0;i<4;i++) res[i] = 0.;}
	~FieldGrad() {}
};

class FieldValueMoving {
public:	
	VR::Ireal res;
	VR::Vector location;
	VR::Ireal size2;	
	VR::Ireal (*function)(VR::Ireal, VR::Ireal);
	
	virtual void addPoints(VR::Vector* cloud, int count);
	virtual void addPoints(VR::Vector* cloud, VR::Ireal* radii, VR::Ireal* sizesLin, int count);
	FieldValueMoving() {}
	FieldValueMoving(VR::Vector location, VR::Ireal (*function)(VR::Ireal, VR::Ireal), VR::real size2) {
		this->location=location; this->function=function; this->size2=size2; res=0.;
	}
	~FieldValueMoving() {}
};

class FieldGradMoving {
public:
	VR::Ireal res[4];
	VR::Vector location[4];
	VR::Ireal size2;	
	VR::Ireal (*function)(VR::Ireal, VR::Ireal);

	virtual void addPoints(VR::Vector* cloud, int count);
	virtual void addPoints(VR::Vector* cloud, VR::Ireal* radii, VR::Ireal* sizesLin, int count);
	FieldGradMoving() {for (int i=0;i<4;i++) res[i] = 0.;}
	~FieldGradMoving() {}
};

/*
	Average field value computations, compute average position for metaballs and compute its field
*/

class FieldValueAvg: public FieldValue {
public:	
	VR::Vector totalP;
	VR::Ireal totalW;
	VR::Ireal totalR;
	void addPoints(int* points, int count);
	FieldValueAvg(VR::Vector* pointCloud, VR::Ireal* radii, VR::Vector location, VR::Ireal (*function)(VR::Ireal, VR::Ireal), VR::real size2) {
		this->pointCloud=pointCloud; this->radii=radii; this->location=location; this->function=function;
		this->size2=size2; res=0.; totalW=0.; totalR=0.; totalP=VR::Vector(0,0,0);
	}
};

class FieldGradAvg: public FieldGrad {
public:
	VR::Vector totalP[4];
	VR::Ireal totalW[4];
	VR::Ireal totalR[4];
	void addPoints(int* points, int count);
	FieldGradAvg() {
		for (int i=0;i<4;i++) {res[i] = 0.; totalW[i]=0.; totalR[i]=0.; totalP[i]=VR::Vector(0,0,0);}
	}
};

class FieldValueMovingAvg: public FieldValueMoving {
public:	
	VR::Vector totalP;
	VR::Ireal totalW;	
	VR::Ireal totalR;
	void addPoints(VR::Vector* cloud, int count);
	void addPoints(VR::Vector* cloud, VR::Ireal* radii, VR::Ireal* sizesLin, int count);
	FieldValueMovingAvg(VR::Vector location, VR::Ireal (*function)(VR::Ireal, VR::Ireal), VR::real size2) {
		this->location=location; this->function=function; this->size2=size2; res=0.; totalW=0.; totalR=0.; totalP=VR::Vector(0,0,0);
	}
};

class FieldGradMovingAvg: public FieldGradMoving {
public:
	VR::Vector totalP[4];
	VR::Ireal totalW[4];
	VR::Ireal totalR[4];
	void addPoints(VR::Vector* cloud, int count);
	void addPoints(VR::Vector* cloud, VR::Ireal* radii, VR::Ireal* sizesLin, int count);
	FieldGradMovingAvg() {
		for (int i=0;i<4;i++) {res[i] = 0.; totalW[i]=0.; totalR[i]=0.; totalP[i]=VR::Vector(0,0,0);}
	}
};