/*
	Copyright 2009 Jerry Ylilammi
*/

#include "readPointCloud.h"

static int compareParticleID(const void* particle1, const void* particle2) {
	return ((Particle*)particle1)->id - ((Particle*)particle2)->id;
}

/*
	This is modified from pflow if I remember correctly, basicly it turns any
	object into mesh thats easy to handle if possible.
*/
static TriObject* GetTriObjectFromNode(INode *iNode, const TimeValue t, bool &deleteIt) {
	deleteIt = false;
	if (iNode == NULL) return NULL;
	Object *obj = iNode->EvalWorldState(t).obj;	
	if (obj == NULL) return NULL;
    if (obj->SuperClassID()==GEOMOBJECT_CLASS_ID) {
		if (obj->IsSubClassOf(triObjectClassID)) {
		  return (TriObject*)obj;
		} else if (obj->CanConvertToType(Class_ID(TRIOBJ_CLASS_ID, 0))) { 
			TriObject *tri = (TriObject *) obj->ConvertToType(t, Class_ID(TRIOBJ_CLASS_ID, 0));
			if (obj != tri) deleteIt = true;
			return tri;
		} else {
			return NULL;
		}
	} else {
		return NULL;
	}
}

/*
	Reads vertices from mesh as point cloud
*/
bool readPointCloud::readTriMesh(INode *node, TimeValue time) {
	bool deleteIt;
	TriObject *triObj = GetTriObjectFromNode(node, time, deleteIt);	
	if (triObj) { // Meshable, lets add the points
		Mesh theMesh = triObj->GetMesh();
		Matrix3 tm = node->GetObjectTM(time);
		Point3* pos;
		for (int i=0; i<theMesh.numVerts;i++) {
			pos = &tm.PointTransform(theMesh.verts[i]);
			posTab.Append(1,&VR::Vector(pos->x, pos->y, pos->z),500);
			if (useRadii)
				radTab.Append(1,&defaultRadius,500);
		}			
		if (deleteIt) triObj->DeleteMe();			
		return true;
	}
	return false; // Not valid tri mesh
}

bool readPointCloud::readTriMesh(INode *node, TimeValue time[2]) {
	bool deleteIt;
	bool deleteIt2;
	TriObject *triObj = GetTriObjectFromNode(node, time[0], deleteIt);	
	if (triObj) { // Meshable, lets add the points
		TriObject *triObj2 = GetTriObjectFromNode(node, time[1], deleteIt2);	
		Mesh theMesh = triObj->GetMesh();
		Mesh theMesh2 = triObj2->GetMesh();			
		Matrix3 tm = node->GetObjectTM(time[0]);
		Matrix3 tm2 = node->GetObjectTM(time[1]);
		Point3* pos;
		Point3* pos2;
		Point3* vel;

		if (theMesh.numVerts != theMesh2.numVerts)
			return false; // Changing vertex count, we can't compute velocities

		for (int i=0; i<theMesh.numVerts;i++) {
			pos = &tm.PointTransform(theMesh.verts[i]);
			pos2 = &tm2.PointTransform(theMesh2.verts[i]);
			posTab.Append(1,&VR::Vector(pos->x, pos->y, pos->z),500);
			vel = &(*pos2 - *pos);
			velTab.Append(1,&VR::Vector(vel->x, vel->y, vel->z),500);

			if (useRadii) {			
				radTab.Append(1,&defaultRadius,500);
				chaTab.Append(1,&zero,500);
			}
		}			
		if (deleteIt) triObj->DeleteMe();			
		return true;
	}
	return false; // Not valid tri mesh
}

/*
	Reads simple particle system
*/
bool readPointCloud::readSimpleParticle(INode *node, TimeValue time[2]) {
	ObjectState tos =  node->EvalWorldState(time[0],TRUE);
	SimpleParticle *pobj = (SimpleParticle*) tos.obj->GetInterface(I_SIMPLEPARTICLEOBJ);
	if (pobj) { // Simple particle system
		pobj->UpdateParticles(time[0], node);
		int count = pobj->parts.Count();
		for (int pid = 0; pid < count; pid++) {
			TimeValue age  = pobj->ParticleAge(time[0],pid);
			if (age!=-1) {
				Point3* pos;
				pos = &pobj->parts.points[pid];			
				posTab.Append(1,&VR::Vector(pos->x, pos->y, pos->z),500);
				Point3* vel;
				vel = &pobj->parts.vels[pid];			
				velTab.Append(1,&(VR::Vector(vel->x, vel->y, vel->z) * (float)(time[1]-time[0])),500);				
				if (useRadii) {
					float rad;
					if (pobj->parts.radius.Count()) rad = pobj->parts.radius[pid];
					radTab.Append(1,&rad,500);
					chaTab.Append(1,&zero,500); // TODO: Evaluate particles at time[1] to get the change in radius
				}
			}
		}
		return true;
	}
	return false;
}

bool readPointCloud::readSimpleParticle(INode *node, TimeValue time) {
	ObjectState tos =  node->EvalWorldState(time,TRUE);
	SimpleParticle *pobj = (SimpleParticle*) tos.obj->GetInterface(I_SIMPLEPARTICLEOBJ);
	if (pobj) { // Simple particle system
		pobj->UpdateParticles(time, node);
		int count = pobj->parts.Count();
		for (int pid = 0; pid < count; pid++) {
			TimeValue age  = pobj->ParticleAge(time,pid);
			if (age!=-1) {
				Point3* pos;
				pos = &pobj->parts.points[pid];			
				posTab.Append(1,&VR::Vector(pos->x, pos->y, pos->z),500);
				if (useRadii) {
					float rad;
					if (pobj->parts.radius.Count()) rad = pobj->parts.radius[pid];
					radTab.Append(1,&rad,500);				
				}
			}
		}
		return true;
	}
	return false;
}

/*
	Reads extended particle system
*/
static void collectParticleData(IParticleObjectExt* epobj, Tab<INode *> &pfList, Tab<Particle> &atTime, BOOL useAllPFEvents, bool useRadii) {
	int count = epobj->NumParticles();
	for (int pid = 0; pid < count; pid++) {
		TimeValue age  = epobj->GetParticleAgeByIndex(pid);
		if (age==-1) continue; // Bad age skip particle
		Point3 *pos = epobj->GetParticlePositionByIndex(pid);
		float scale = epobj->GetParticleScaleByIndex(pid); 
		INode *group = epobj->GetParticleGroup(pid);
		BOOL useParticle = TRUE;			
		if (!useAllPFEvents) {
			useParticle = FALSE;
			for (int j = 0; j < pfList.Count(); j++)
				if (group == pfList[j]) {
					useParticle = TRUE;
					break;
				}
		}
		if ((pos) && (useParticle)) {				
			Particle p;
			p.id = epobj->GetParticleBornIndex(pid);
			p.position = VR::Vector(pos->x, pos->y, pos->z);			
			p.radius = scale;
			atTime.Append(1,&p,500);
		}
	}
}

bool readPointCloud::readExtParticle(INode *node, IParamBlock2 *pblock, TimeValue time) {
	ObjectState tos =  node->EvalWorldState(time,TRUE);
	IParticleObjectExt* epobj = (IParticleObjectExt*) tos.obj->GetInterface(PARTICLEOBJECTEXT_INTERFACE);
	if (epobj) {
		Tab<INode *> pfList;
		BOOL useAllPFEvents;
		pblock->GetValue(pb_allPflow,0,useAllPFEvents,FOREVER);
		int pfeventCount = pblock->Count(pb_pflow);
		for (int i = 0; i < pfeventCount; i++) {
			INode *pfevent;
			pblock->GetValue(pb_pflow, 0, pfevent, FOREVER, i);
			if (node)
				pfList.Append(1,&pfevent);
		}
		
		Tab<Particle> atTime0;		
		epobj->UpdateParticles(node, time);
		collectParticleData(epobj, pfList, atTime0, useAllPFEvents, useRadii);

		int count = atTime0.Count(); 
		for (int i=0; i<count; i++) {			
			posTab.Append(1,&(atTime0[i].position),500);
			if (useRadii)
				radTab.Append(1,&(atTime0[i].radius),500);							
		}

		atTime0.SetCount(0);
		return true;
	}
	return false;
}

bool readPointCloud::readExtParticle(INode *node, IParamBlock2 *pblock, TimeValue time[2]) {
	ObjectState tos =  node->EvalWorldState(time[0],TRUE);
	IParticleObjectExt* epobj = (IParticleObjectExt*) tos.obj->GetInterface(PARTICLEOBJECTEXT_INTERFACE);
	if (epobj) {
		Tab<INode *> pfList;
		BOOL useAllPFEvents;
		pblock->GetValue(pb_allPflow,0,useAllPFEvents,FOREVER);
		int pfeventCount = pblock->Count(pb_pflow);
		for (int i = 0; i < pfeventCount; i++) {
			INode *pfevent;
			pblock->GetValue(pb_pflow, 0, pfevent, FOREVER, i);
			if (node)
				pfList.Append(1,&pfevent);
		}

		// Collect particle data at the start and end of the time interval
		Tab<Particle> atTime0;		
		epobj->UpdateParticles(node, time[0]);
		collectParticleData(epobj, pfList, atTime0, useAllPFEvents, useRadii);
		//thread_local(current_stdout)->printf("Count0 %d\n", atTime0.Count());

		Tab<Particle> atTime1;
		epobj->UpdateParticles(node, time[1]);
		collectParticleData(epobj, pfList, atTime1, useAllPFEvents, useRadii);
		//thread_local(current_stdout)->printf("Count1 %d\n", atTime1.Count());

		// Intersect two particle data states, only those particles that exists in both are used, those born and deleted within are ignored
		atTime0.Sort(compareParticleID);
		atTime1.Sort(compareParticleID);

		int count = atTime0.Count(); 
		int j = 0;
		for (int i=0; i<count; i++) {			
			while (atTime1[j].id < atTime0[i].id)
				j++;
			if (atTime0[i].id == atTime1[j].id) { // State at time 0 and time 1 both exists, append particle to our Tabs				
				posTab.Append(1,&(atTime0[i].position),500);								
				VR::Vector velocity = atTime1[j].position-atTime0[i].position; 				
				velTab.Append(1,&velocity,500);				
				if (useRadii) {
					radTab.Append(1,&(atTime0[i].radius),500);
					float change = atTime1[j].radius-atTime0[i].radius;				
					chaTab.Append(1,&change,500);				
				}				
			}			
		}

		atTime0.SetCount(0);
		atTime1.SetCount(0);
		return true;
	}
	return false;
}	

/*
	Reads all nodes into points
*/

void readPointCloud::read(bool useRadii, int &pointCount, VR::Vector* &position, VR::Ireal* &radii, IParamBlock2 *pblock, TimeValue time) {
	this->useRadii = useRadii;

	// Start parsing nodes
	int nodeCount = pblock->Count(pb_nodes);
	for (int i=0; i<nodeCount; i++) {
		INode *node;
		pblock->GetValue(pb_nodes,time,node,Interval(time,time),i);
		if (!node) continue; // No node skip

		// Mesh
		if (readTriMesh(node, time))
			continue;

		// Particles
		ObjectState tos =  node->EvalWorldState(time,TRUE);
		if (tos.obj->IsParticleSystem()) {				
			// Simple particle system
			if (readSimpleParticle(node, time))
				continue;	
			// Extended particle system
			if (readExtParticle(node, pblock, time))
				continue;	
		}		

		// TODO: Add support for more objects here
	}

	// Extra metaball points, the Point3 array which can be used through maxscript for example
	int count = pblock->Count(pb_points);
	int count2 = pblock->Count(pb_pointSizes);
	Interval ivalid = Interval(time,time);
	for (int i=0; i<count; i++) {
		Point3 pos;
		pblock->GetValue(pb_points,time,pos,ivalid,i);
		posTab.Append(1,&VR::Vector(pos.x, pos.y, pos.z),500);
		if (useRadii) {		
			if (i < count2) {
				float size;
				pblock->GetValue(pb_pointSizes,time,size,ivalid,i);
				radTab.Append(1,&size,500);
			} else {
				radTab.Append(1,&defaultRadius,500);
			}
		}
	}

	// Convert tabs to arrays
	pointCount = posTab.Count();

	position = new VR::Vector[pointCount]; 	
	for (int i=0;i<pointCount;i++)
		position[i] = posTab[i];
	posTab.SetCount(0); // Free memory

	if (useRadii) {
		radii = new VR::Ireal[pointCount];
		for (int i=0;i<pointCount;i++)
			radii[i] = radTab[i];
		radTab.SetCount(0); // Free memory		
	}	
}

void readPointCloud::read(bool useRadii, int &pointCount, VR::Vector* &position, VR::Vector* &velocity, VR::Ireal* &radii, VR::Ireal* &radiiChange, IParamBlock2 *pblock, TimeValue time[2]) {
	this->useRadii = useRadii;

	// Start parsing nodes
	int nodeCount = pblock->Count(pb_nodes);
	for (int i=0; i<nodeCount; i++) {
		INode *node;
		pblock->GetValue(pb_nodes,time[0],node,Interval(time[0],time[0]),i);
		if (!node) continue; // No node skip

		// Mesh
		if (readTriMesh(node, time))
			continue;

		// Particles
		ObjectState tos =  node->EvalWorldState(time[0],TRUE);
		if (tos.obj->IsParticleSystem()) {				
			// Simple particle system
			if (readSimpleParticle(node, time))
				continue;	
			// Extended particle system
			if (readExtParticle(node, pblock, time))
				continue;	
		}		

		// TODO: Add support for more objects here
	}

	// Extra metaball points, the Point3 array which can be used through maxscript for example
	int count = pblock->Count(pb_points);
	int count2 = pblock->Count(pb_pointSizes);
	Interval ivalid = Interval(time[0],time[0]); // This doesn't matter, we would get the interval where the position is valid through it but we don't need it
	for (int i=0; i<count; i++) {
		Point3 pos;
		pblock->GetValue(pb_points,time[0],pos,ivalid,i);
		posTab.Append(1,&VR::Vector(pos.x, pos.y, pos.z),500);
		Point3 pos2;
		pblock->GetValue(pb_points,time[1],pos2,ivalid,i);
		pos = pos2 - pos;
		velTab.Append(1,&VR::Vector(pos.x, pos.y, pos.z),500);			
		if (useRadii) {		
			if (i < count2) {
				float size;
				pblock->GetValue(pb_pointSizes,time[0],size,ivalid,i);
				radTab.Append(1,&size,500);
				float size2;
				pblock->GetValue(pb_pointSizes,time[1],size2,ivalid,i);
				size = size2 - size;
				chaTab.Append(1,&size,500);
			} else {
				radTab.Append(1,&defaultRadius,500);
				chaTab.Append(1,&zero,500);
			}
		}
	}

	// Convert tabs to arrays
	pointCount = posTab.Count();

	position = new VR::Vector[pointCount]; 	
	for (int i=0;i<pointCount;i++)
		position[i] = posTab[i];
	posTab.SetCount(0); // Free memory

	velocity = new VR::Vector[pointCount];
	for (int i=0;i<pointCount;i++)
		velocity[i] = velTab[i];
	velTab.SetCount(0); // Free memory

	if (useRadii) {
		radii = new VR::Ireal[pointCount];
		for (int i=0;i<pointCount;i++)
			radii[i] = radTab[i];
		radTab.SetCount(0); // Free memory
		
		radiiChange = new VR::Ireal[pointCount];
		for (int i=0;i<pointCount;i++)
			radiiChange[i] = chaTab[i];
		chaTab.SetCount(0); // Free memory				
	}	
}